#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""ADS-B Cursor-On-Target Class Definitions."""


import asyncio
import logging

import aiohttp
import pytak

import adsbcot

# We won't use pyModeS if it isn't installed:
try:
    import pyModeS.streamer.source
    import pyModeS.streamer.decode
    import pyModeS as pms
except ImportError:
    pass


__author__ = "Greg Albrecht W2GMD <oss@undef.net>"
__copyright__ = "Copyright 2022 Greg Albrecht"
__license__ = "Apache License, Version 2.0"


class ADSBWorker(pytak.MessageWorker):

    """Reads ADS-B Data from inputs, renders to CoT, and puts on queue."""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        event_queue,
        url,
        cot_stale,
        poll_interval: int = None,
        filters: dict = None,
    ):
        super().__init__(event_queue, url)
        self.poll_interval: int = int(poll_interval or adsbcot.DEFAULT_POLL_INTERVAL)
        self.url = url
        self.cot_stale = cot_stale
        self.filters = filters
        self._logger.debug("cot_stale='%s'", self.cot_stale)

    async def handle_message(self, aircraft: list) -> None:
        """
        Transforms Aircraft AIS data to CoT and puts it onto tx queue.
        """
        if not isinstance(aircraft, list):
            self._logger.warning("Invalid aircraft data, should be a Python list.")
            return False

        if not aircraft:
            self._logger.warning("Empty aircraft list")
            return False

        i = 1
        for craft in aircraft:
            self._logger.debug("craft='%s'", craft)

            if self.filters:
                if "ICAO" in self.filters:
                    icao = craft.get("hex", "").upper()
                    if icao not in self.filters.get("ICAO", "include"):
                        continue
                    if icao in self.filters.get("ICAO", "exclude"):
                        continue
                elif "FLIGHT" in self.filters:
                    flight = craft.get("flight", "").upper()
                    if flight not in self.filters.get("FLIGHT", "include"):
                        continue
                    if flight in self.filters.get("FLIGHT", "exclude"):
                        continue

            event = adsbcot.adsb_to_cot(craft, stale=self.cot_stale)
            if not event:
                self._logger.debug("Empty CoT Event for craft=%s", craft)
                i += 1
                continue

            self._logger.debug(
                "Handling %s/%s ICAO: %s Flight: %s Category: %s Registration: %s",
                i,
                len(aircraft),
                craft.get("hex"),
                craft.get("flight"),
                craft.get("category"),
                craft.get("reg"),
            )
            await self._put_event_queue(event)
            i += 1

    async def _get_dump1090_feed(self):
        """
        Polls the dump1090 JSON API and passes data to message handler.
        """
        async with aiohttp.ClientSession() as session:
            resp = await session.request(method="GET", url=self.url.geturl())
            resp.raise_for_status()

            json_resp = await resp.json()
            aircraft = json_resp.get("aircraft")
            self._logger.debug(
                "Retrieved %s aircraft from %s", len(aircraft), self.url.geturl()
            )

            await self.handle_message(aircraft)

    async def run(self, number_of_iterations=-1):
        """Runs this Thread, Reads from Pollers."""
        self._logger.info("Running ADSBWorker with URL '%s'", self.url.geturl())

        while 1:
            await self._get_dump1090_feed()
            await asyncio.sleep(self.poll_interval)


class ADSBNetWorker(ADSBWorker):
    """Reads ADS-B Data from network, renders to COT, and puts on queue."""

    def __init__(
        self, event_queue, net_queue, data_type, cot_stale, filters
    ):  # NOQA pylint: disable=too-many-arguments
        super().__init__(event_queue, None, cot_stale, None, filters)
        self.net_queue = net_queue
        self.data_type = data_type
        self.local_buffer_adsb_msg = []
        self.local_buffer_adsb_ts = []
        self.local_buffer_commb_msg = []
        self.local_buffer_commb_ts = []

    def _reset_local_buffer(self):
        """Resets Socket Buffers."""
        self.local_buffer_adsb_msg = []
        self.local_buffer_adsb_ts = []
        self.local_buffer_commb_msg = []
        self.local_buffer_commb_ts = []

    async def run(
        self, number_of_iterations=-1
    ):  # NOQA pylint: disable=too-many-locals, too-many-branches
        """Runs the main process loop."""
        self._logger.info("Running ADSBNetWorker for data_type='%s'", self.data_type)

        self._reset_local_buffer()

        decoder = pyModeS.streamer.decode.Decode()
        net_client = pyModeS.streamer.source.NetSource("x", 1, self.data_type)

        while 1:
            messages = []
            received = await self.net_queue.get()
            if not received:
                continue

            net_client.buffer.extend(received)
            if "beast" in self.data_type:
                messages = net_client.read_beast_buffer()
            elif "raw" in self.data_type:
                messages = net_client.read_raw_buffer()
            elif "skysense" in self.data_type:
                messages = net_client.read_skysense_buffer()

            self._logger.debug("Received %s messages", len(messages))

            if not messages:
                continue

            for msg, t_msg in messages:
                if len(msg) != 28:  # wrong data length
                    continue

                dl_fmt = pms.df(msg)

                if dl_fmt != 17:  # not ADSB
                    continue

                if pms.crc(msg) != 0:  # CRC fail
                    continue

                # icao = pms.adsb.icao(msg)
                # typecode = pms.adsb.typecode(msg)

                if dl_fmt in (17, 18):
                    self.local_buffer_adsb_msg.append(msg)
                    self.local_buffer_adsb_ts.append(t_msg)
                elif dl_fmt in (20, 21):
                    self.local_buffer_commb_msg.append(msg)
                    self.local_buffer_commb_ts.append(t_msg)
                else:
                    continue

            if len(self.local_buffer_adsb_msg) > 1:
                decoder.process_raw(
                    self.local_buffer_adsb_ts,
                    self.local_buffer_adsb_msg,
                    self.local_buffer_commb_ts,
                    self.local_buffer_commb_msg,
                )
                self._reset_local_buffer()

            acs = decoder.get_aircraft()
            for key, val in acs.items():
                # self._logger.debug("acs=%s", acs[k])
                lat = val.get("lat")
                lon = val.get("lon")
                flight = val.get("call", key)
                alt_geom = val.get("alt")
                gnds = val.get("gs")
                reg = val.get("r")
                trk = val.get("trk")
                if lat and lon and flight and alt_geom and gnds and trk:
                    aircraft = [
                        {
                            "hex": key,
                            "lat": lat,
                            "lon": lon,
                            "flight": flight.replace("_", ""),
                            "alt_geom": alt_geom,
                            "gs": gnds,
                            "reg": reg,
                            "trk": trk,
                        }
                    ]
                    await self.handle_message(aircraft)
                else:
                    continue


class ADSBNetReceiver:  # pylint: disable=too-few-public-methods
    """Reads ADS-B Data from network and puts on queue."""

    _logger = logging.getLogger(__name__)
    if not _logger.handlers:
        _logger.setLevel(adsbcot.LOG_LEVEL)
        _console_handler = logging.StreamHandler()
        _console_handler.setLevel(adsbcot.LOG_LEVEL)
        _console_handler.setFormatter(adsbcot.LOG_FORMAT)
        _logger.addHandler(_console_handler)
        _logger.propagate = False
    logging.getLogger("asyncio").setLevel(adsbcot.LOG_LEVEL)

    def __init__(self, net_queue, url, data_type):
        self.net_queue = net_queue
        self.url = url
        self.data_type = data_type

    async def run(self):
        """Runs the main process loop."""
        self._logger.info("Running ADSBNetReceiver for URL=%s", self.url.geturl())

        if ":" in self.url.path:
            host, port = self.url.path.split(":")
        else:
            host = self.url.path
            if self.data_type == "raw":
                port = adsbcot.DEFAULT_DUMP1090_TCP_RAW_PORT
            elif self.data_type == "beast":
                port = adsbcot.DEFAULT_DUMP1090_TCP_BEAST_PORT
            else:
                raise Exception("Invalid data_type='%s'" % self.data_type)

        self._logger.debug("host=%s port=%s", host, port)

        reader, _ = await asyncio.open_connection(host, port)

        if self.data_type == "raw":
            while 1:
                received = await reader.readline()
                self.net_queue.put_nowait(received)
        elif self.data_type == "beast":
            while 1:
                received = await reader.read(4096)
                self.net_queue.put_nowait(received)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""ADS-B Cursor-on-Target Constants."""

import logging
import os

__author__ = "Greg Albrecht W2GMD <oss@undef.net>"
__copyright__ = "Copyright 2022 Greg Albrecht"
__license__ = "Apache License, Version 2.0"


if os.getenv("DEBUG"):
    LOG_LEVEL = logging.DEBUG
    LOG_FORMAT = logging.Formatter(
        (
            "%(asctime)s adsbcot %(levelname)s %(name)s.%(funcName)s:%(lineno)d "
            " - %(message)s"
        )
    )
    logging.debug("adsbcot Debugging Enabled via DEBUG Environment Variable.")
else:
    LOG_LEVEL = logging.INFO
    LOG_FORMAT = logging.Formatter(("%(asctime)s adsbcot %(levelname)s - %(message)s"))

DEFAULT_POLL_INTERVAL: int = 30
DEFAULT_COT_STALE: int = 120
DEFAULT_DUMP1090_TCP_RAW_PORT: str = 30002
DEFAULT_DUMP1090_TCP_BEAST_PORT: str = 30005



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""ADS-B Cursor-on-Target Gateway Commands."""

import argparse
import asyncio
import configparser
import collections
import logging
import os
import platform
import sys
import typing

from urllib.parse import ParseResult, urlparse

import pytak

import adsbcot

# We won't use pyModeS if it isn't installed:
WITH_PYMODES = False
try:
    import pyModeS  # pylint: disable=unused-import

    WITH_PYMODES = True
except ImportError:
    pass

# Python 3.6 support:
if sys.version_info[:2] >= (3, 7):
    from asyncio import get_running_loop
else:
    from asyncio import _get_running_loop as get_running_loop

__author__ = "Greg Albrecht W2GMD <oss@undef.net>"
__copyright__ = "Copyright 2022 Greg Albrecht"
__license__ = "Apache License, Version 2.0"


async def main(  # pylint: disable=too-many-locals
    opts: typing.ChainMap[str, typing.Any]
) -> None:  # pylint: disable=too-many-locals
    tx_queue: asyncio.Queue = asyncio.Queue()
    rx_queue: asyncio.Queue = asyncio.Queue()
    cot_url: ParseResult = urlparse(opts.get("COT_URL"))

    # Create our COT Event Queue Worker
    reader, writer = await pytak.protocol_factory(cot_url)
    write_worker = pytak.EventTransmitter(tx_queue, writer)
    read_worker = pytak.EventReceiver(rx_queue, reader)

    dump1090_url: ParseResult = urlparse(opts.get("DUMP1090_URL"))

    tasks = set()

    # ADSB Workers (receivers):
    if "http" in dump1090_url.scheme:
        message_worker = adsbcot.ADSBWorker(
            event_queue=tx_queue,
            url=dump1090_url,
            poll_interval=opts.get("POLL_INTERVAL"),
            cot_stale=opts.get("COT_STALE"),
            filters=opts.get("FILTERS"),
        )
    elif "tcp" in dump1090_url.scheme:
        if not WITH_PYMODES:
            print("ERROR from adsbcot")
            print("Please reinstall adsbcot with pyModeS support: ")
            print("$ python3 -m pip install adsbcot[with_pymodes]")
            print("Exiting...")
            raise Exception

        net_queue: asyncio.Queue = asyncio.Queue()

        if "+" in dump1090_url.scheme:
            _, data_type = dump1090_url.scheme.split("+")
        else:
            data_type = "raw"

        adsbnetreceiver = adsbcot.ADSBNetReceiver(net_queue, dump1090_url, data_type)

        tasks.add(asyncio.ensure_future(adsbnetreceiver.run()))

        message_worker = adsbcot.ADSBNetWorker(
            event_queue=tx_queue,
            net_queue=net_queue,
            data_type=data_type,
            cot_stale=opts.get("COT_STALE"),
            filters=opts.get("FILTERS"),
        )

    await tx_queue.put(pytak.hello_event(f"adsbxcot@{platform.node()}"))

    tasks.add(asyncio.ensure_future(message_worker.run()))
    tasks.add(asyncio.ensure_future(read_worker.run()))
    tasks.add(asyncio.ensure_future(write_worker.run()))

    done, _ = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)

    for task in done:
        print(f"Task completed: {task}")


def cli() -> None:
    """Command Line interface for ADS-B Cursor-on-Target Gateway."""

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-c", "--CONFIG_FILE", dest="CONFIG_FILE", default="config.ini", type=str
    )

    namespace = parser.parse_args()
    cli_args = {k: v for k, v in vars(namespace).items() if v is not None}

    # Read config file:
    config_file = cli_args.get("CONFIG_FILE")
    logging.info("Reading configuration from %s", config_file)
    config = configparser.ConfigParser()
    config.read(config_file)

    # Combined command-line args with config file:
    combined_config = collections.ChainMap(cli_args, os.environ, config["adsbcot"])

    if combined_config.get("FILTER_CONFIG"):
        filter_config = combined_config.get("FILTER_CONFIG")
        logging.info("Reading FILTER_CONFIG from %s", filter_config)
        filters = configparser.ConfigParser()
        filters.read(filter_config)
        combined_config = collections.ChainMap(combined_config, {"FILTERS": filters})

    if sys.version_info[:2] >= (3, 7):
        asyncio.run(main(combined_config), debug=os.getenv("DEBUG"))
    else:
        loop = get_running_loop()
        try:
            loop.run_until_complete(main(combined_config))
        finally:
            loop.close()


if __name__ == "__main__":
    cli()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ADS-B Cursor-On-Target Gateway.

"""
ADS-B Cursor-On-Target Gateway.
~~~~


:author: Greg Albrecht W2GMD <oss@undef.net>
:copyright: Copyright 2022 Greg Albrecht
:license: Apache License, Version 2.0
:source: <https://github.com/ampledata/adsbcot>

"""

from .constants import (  # NOQA
    LOG_FORMAT,
    LOG_LEVEL,
    DEFAULT_POLL_INTERVAL,
    DEFAULT_COT_STALE,
    DEFAULT_DUMP1090_TCP_RAW_PORT,
    DEFAULT_DUMP1090_TCP_BEAST_PORT,
)

from .functions import adsb_to_cot  # NOQA

from .classes import ADSBWorker, ADSBNetReceiver, ADSBNetWorker  # NOQA

__author__ = "Greg Albrecht W2GMD <oss@undef.net>"
__copyright__ = "Copyright 2022 Greg Albrecht"
__license__ = "Apache License, Version 2.0"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""ADS-B Cursor-on-Target Gateway Functions."""

import datetime
import os
import platform

import xml.etree.ElementTree as ET

import aircot
import pytak

import adsbcot

__author__ = "Greg Albrecht W2GMD <oss@undef.net>"
__copyright__ = "Copyright 2022 Greg Albrecht"
__license__ = "Apache License, Version 2.0"


def adsb_to_cot_xml(  # pylint: disable=too-many-locals,too-many-statements
    craft: dict, cot_type: str = None, stale: int = None
) -> ET.Element:
    """
    Transforms a Dump1090 ADS-B Aircraft Object to a Cursor-on-Target PLI XML.
    """
    time = datetime.datetime.now(datetime.timezone.utc)
    cot_stale = stale or adsbcot.DEFAULT_COT_STALE

    lat = craft.get("lat")
    lon = craft.get("lon")

    if lat is None or lon is None:
        return None

    icao_hex = craft.get("hex", "").upper()
    if not icao_hex:
        return None

    name = f"ICAO-{icao_hex}"

    flight = craft.get("flight", "").strip()
    if flight:
        callsign = flight
    else:
        callsign = icao_hex

    cot_type = aircot.adsb_to_cot_type(craft.get("hex"), craft.get("category"), flight)

    point: ET.Element = ET.Element("point")
    point.set("lat", str(lat))
    point.set("lon", str(lon))
    point.set("ce", str(craft.get("nac_p", "9999999.0")))
    point.set("le", str(craft.get("nac_v", "9999999.0")))

    # alt_geom: geometric (GNSS / INS) altitude in feet referenced to the
    #           WGS84 ellipsoid
    alt_geom = int(craft.get("alt_geom", 0))
    if alt_geom:
        point.set("hae", str(alt_geom * 0.3048))
    else:
        point.set("hae", "9999999.0")

    uid: ET.Element = ET.Element("UID")
    uid.set("Droid", name)

    contact: ET.Element = ET.Element("contact")
    contact.set("callsign", callsign)

    track: ET.Element = ET.Element("track")
    track.set("course", str(craft.get("trk", "9999999.0")))

    # gs: ground speed in knots
    gnds = int(craft.get("gs", 0))
    if gnds:
        track.set("speed", str(gnds * 0.514444))
    else:
        track.set("speed", "9999999.0")

    detail = ET.Element("detail")
    detail.set("uid", name)
    detail.append(uid)
    detail.append(contact)
    detail.append(track)

    remarks = ET.Element("remarks")
    _remarks = f"Squawk: {craft.get('squawk')} Category: {craft.get('category')}"

    if flight:
        _remarks = f"{icao_hex}({flight}) {_remarks}"
    else:
        _remarks = f"{icao_hex} {_remarks}"

    if bool(os.environ.get("DEBUG")):
        _remarks = f"{_remarks} via (via adsbxcot@{platform.node()})"

    detail.set("remarks", _remarks)
    remarks.text = _remarks
    detail.append(remarks)

    root = ET.Element("event")
    root.set("version", "2.0")
    root.set("type", cot_type)
    root.set("uid", name)
    root.set("how", "m-g")
    root.set("time", time.strftime(pytak.ISO_8601_UTC))
    root.set("start", time.strftime(pytak.ISO_8601_UTC))
    root.set(
        "stale",
        (time + datetime.timedelta(seconds=int(cot_stale))).strftime(
            pytak.ISO_8601_UTC
        ),
    )
    root.append(point)
    root.append(detail)

    return root


def adsb_to_cot(craft: dict, cot_type: str = None, stale: int = None) -> bytes:
    """
    Transforms a Dump1090 ADS-B Aircraft Object to a Cursor-on-Target PLI String.
    """
    return ET.tostring(adsb_to_cot_xml(craft, cot_type, stale))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""ADSBCOT Setup.
Source:: https://github.com/ampledata/adsbcot
"""

import os
import sys

import setuptools

__title__ = "adsbcot"
__version__ = "4.0.1"
__author__ = "Greg Albrecht W2GMD <oss@undef.net>"
__copyright__ = "Copyright 2022 Greg Albrecht"
__license__ = "Apache License, Version 2.0"


def publish():
    """Function for publishing package to pypi."""
    if sys.argv[-1] == "publish":
        os.system("python setup.py sdist")
        os.system("twine upload dist/*")
        sys.exit()


publish()


setuptools.setup(
    version=__version__,
    name=__title__,
    packages=[__title__],
    package_dir={__title__: __title__},
    url=f"https://github.com/ampledata/{__title__}",
    description="ADS-B to Cursor-On-Target Gateway.",
    author="Greg Albrecht",
    author_email="oss@undef.net",
    package_data={"": ["LICENSE"]},
    license="Apache License, Version 2.0",
    long_description=open("README.rst").read(),
    zip_safe=False,
    include_package_data=True,
    install_requires=["aircot", "pytak >= 3.1.0", "aiohttp"],
    extras_require={"with_pymodes": "pymodes==2.8"},
    classifiers=[
        "Programming Language :: Python",
        "License :: OSI Approved :: Apache Software License",
    ],
    keywords=["ADS-B", "ADSB", "Cursor on Target", "ATAK", "TAK", "CoT"],
    entry_points={"console_scripts": ["adsbcot = adsbcot.commands:cli"]},
)
